
#include <math.h>
#include <cstdlib>
#include <cstring>
#include <cctype>

#include "lg/interface.h"
#include "lg/scrservices.h"
#include "lg/scrmanagers.h"
#include "lg/scrmsgs.h"
#include "lg/properties.h"
#include "lg/propdefs.h"
#include "lg/input.h"
#include "ScriptLib.h"
#include "tnhScript.h"


/*******************
 * CommandControl
 */
long cScr_CommandCtrl::OnControlString(const char* pszControl, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IDebugScrSrv> pDebugSrv(g_pScriptManager);
	char* pszCommand;
	for (pszCommand = strtok(const_cast<char*>(pszControl), ";"); pszCommand; pszCommand = strtok(NULL, ";"))
	{
		pDebugSrv->Command(pszCommand, "", "", "", "", "", "", "");
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*******************
 * TrapFrobber
 */
long cScr_TrapFrobber::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	sFrobMsg pFrobMsg;
	//sMultiParm mpRet;
	pFrobMsg.from = m_iObjId;
	pFrobMsg.to = m_iObjId;
	pFrobMsg.time = pMsg->time;
	pFrobMsg.flags = 4; // And someday I might even learn what this means.
	pFrobMsg.SrcObjId = m_iObjId;
	pFrobMsg.DstObjId = 0;
	pFrobMsg.Frobber = 0;
	pFrobMsg.DstLoc = kFrobLocNone;
	pFrobMsg.Sec = 0.0;
	pFrobMsg.Abort = 0;
	if (GetObjectParamInt(m_iObjId, "inv"))
	{
		pFrobMsg.message = "FrobInvEnd";
		pFrobMsg.SrcLoc = kFrobLocInv;
	}
	else
	{
		pFrobMsg.message = "FrobWorldEnd";
		pFrobMsg.SrcLoc = kFrobLocWorld;
	}

	g_pScriptManager->SendMessage(&pFrobMsg, pReply);
	return 0;
}


/*******************
 * TrapMoveRelative
 */
long __stdcall cScr_MoveRelativeTrap::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		FixupPlayerLinks();
	}
	return 1;
}

/*
 * cScrVec.pos.x = Bank;
 * cScrVec.pos.y = Pitch;
 * cScrVec.pos.z = Heading;
 */
struct MoveRelative_Struct
{
	cScrVec	vPosition,
			vFacing,
			vVelocity,
			vRotation;
	bool bRotateAxes;
};

int cScr_MoveRelativeTrap::LinkIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	MoveRelative_Struct* pSDelta = reinterpret_cast<MoveRelative_Struct*>(pData);

	sLink sl;
	pLQ->Link(&sl);
	int iTargetObj = sl.dest;

	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	if (pSDelta->bRotateAxes)
	{
		pObjSrv->Teleport(iTargetObj, pSDelta->vPosition, pSDelta->vFacing, iTargetObj);
	}
	else
	{
		cScrVec vDestPos, vDestFacing;
		pObjSrv->Position(vDestPos,iTargetObj);
		vDestPos += pSDelta->vPosition;
		pObjSrv->Facing(vDestFacing,iTargetObj);
		vDestFacing += pSDelta->vFacing;
		pObjSrv->Teleport(iTargetObj, vDestPos, vDestFacing, 0);
	}

	/*
	SService<IPhysSrv> pPS(g_pScriptManager);
	pPS->GetVelocity(iTargetObj,vDestVel);
	vDestVel += pSDelta->vVelocity;
	pPS->SetVelocity(iTargetObj, vDestVel);
	*/
	SInterface<IPropertyManager> pPropMan(g_pScriptManager);
	SInterface<IProperty> pProp;
	pProp = pPropMan->GetPropertyNamed("PhysState");
	if (pProp)
	{
		IProperty* pPhysStateProp;
		sPhysState* pPhysState;
		if (E_NOINTERFACE != pProp->QueryInterface(IID_IPhysStateProperty, reinterpret_cast<void**>(&pPhysStateProp)))
		{
			if (pPhysStateProp->Get(iTargetObj, reinterpret_cast<void**>(&pPhysState)))
			{
				cScrVec v;
				v = pPhysState->vels;
				pPhysState->vels = v + pSDelta->vVelocity;
				v = pPhysState->rot_vels;
				pPhysState->rot_vels = v + pSDelta->vRotation;
				pPhysStateProp->Set(iTargetObj, reinterpret_cast<void*>(pPhysState));
			}
			pPhysStateProp->Release();
		}
	}

	return 1;
}

long cScr_MoveRelativeTrap::OnControlString(const char* pszControl, sScrMsg* pMsg, sMultiParm* pReply)
{
	MoveRelative_Struct vDelta;

	char* pszParams = new char[strlen(pszControl)+1];
	strcpy(pszParams, pszControl);
	char* pszCurrent;
	for (pszCurrent = strtok(pszParams, ";"); pszCurrent; pszCurrent = strtok(NULL, ";"))
	{
		// position/heading or velocity/rotation?
		bool delta;
		if (tolower(*pszCurrent) == 'd')
		{
			delta = TRUE;
			pszCurrent++;
		}
		else
			delta = FALSE;
		// get value
		char* pszValue = strchr(pszCurrent, '=') + 1;
		float value = strtod(pszValue, NULL);
		switch (tolower(*pszCurrent))
		{
		  case 'x':
			if (delta)
				vDelta.vVelocity.pos.x = value;
			else
				vDelta.vPosition.pos.x = value;
			break;
		  case 'y':
			if (delta)
				vDelta.vVelocity.pos.y = value;
			else
				vDelta.vPosition.pos.y = value;
			break;
		  case 'z':
			if (delta)
				vDelta.vVelocity.pos.z = value;
			else
				vDelta.vPosition.pos.z = value;
			break;
		  case 'b':
			if (delta)
				vDelta.vRotation.pos.x = value;
			else
				vDelta.vFacing.pos.x = value;
			break;
		  case 'p':
			if (delta)
				vDelta.vRotation.pos.y = value;
			else
				vDelta.vFacing.pos.y = value;
			break;
		  case 'h':
			if (delta)
				vDelta.vRotation.pos.z = value;
			else
				vDelta.vFacing.pos.z = value;
			break;
		}
	}
	delete[] pszParams;

	vDelta.bRotateAxes = GetObjectParamInt(m_iObjId, "rotate_axes");
	IterateLinks("ControlDevice", m_iObjId, 0, LinkIterFunc, this, reinterpret_cast<void*>(&vDelta));

	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*******************
 * TrapCampaignQVar
 */
long cScr_CampaignQVarTrap::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (!pPropSrv->Possessed(m_iObjId,"TrapQVar"))
		return 1;
	cMultiParm mpQVar;
	pPropSrv->Get(mpQVar, m_iObjId, "TrapQVar", NULL);
	if (mpQVar.type != kMT_String || static_cast<const char*>(mpQVar) == NULL)
		return 1;
	
	/*
	char* pszDup = new char[strlen(static_cast<const char*>(mpQVar))+1];
	strcpy(pszDup, static_cast<const char*>(mpQVar));
	*/
	char* pszDup = mpQVar.psz;
	int iValue = 0;
	int iCurrent = 0;
	char cOperation = *pszDup;
	iValue = strtol(pszDup+1, NULL, 10);
	char *pszName = strchr(pszDup, ':') + 1;
	if (!bTurnOn)
		cOperation &= 0x80;		// flag high-bit for inverse operation
	SService<IQuestSrv> pQSrv(g_pScriptManager);
	if (pQSrv->Exists(pszName))
		iCurrent = pQSrv->Get(pszName);
	switch (cOperation)
	{
	  case '=':
		iCurrent = iValue;
		break;
	  case '!':
	  case '|':
		iCurrent ^= iValue;
		break;
	  case ('!'+0x80):
	  case ('|'+0x80):
		iCurrent &= ~iValue;
		break;
	  case '+':
	  case ('-'+0x80):
		iCurrent += iValue;
		break;
	  case '-':
	  case ('+'+0x80):
		iCurrent -= iValue;
		break;
	  case '*':
	  case ('/'+0x80):
	  case ('%'+0x80):
		iCurrent *= iValue;
		break;
	  case '/':
	  case ('*'+0x80):
		iCurrent /= iValue;
		break;
	  case '%':
		iCurrent %= iValue;
		break;
	  case '{':
	  case ('}'+0x80):
		iCurrent <<= iValue;
		break;
	  case '}':
	  case ('{'+0x80):
		iCurrent >>= iValue;
		break;
	}
	pQSrv->Set(pszName, iCurrent, 1);
	//delete[] pszDup;
	g_pMalloc->Free(mpQVar.psz);
	return 0;
}


/*****************
 * TrapKillEvents
 */
long cScr_KillEvents::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	SInterface<IInputBinder> pInput(g_pScriptManager);
	SService<IDarkInvSrv> pInvSrv(g_pScriptManager);
	if (bTurnOn)
	{
		long lEventMask = 0;
		char* pszParams = GetObjectParamString(m_iObjId, "event_mask");
		if (pszParams)
		{
			if (*pszParams == '$')
			{
				char* pszStart = pszParams + 1;
				char* pszEnd = pszStart;
				while (isalnum(*pszEnd) || (*pszEnd == '_'))
					++pszEnd;
				*pszEnd = '\0';
				if (pszEnd > pszStart)
				{
					SService<IQuestSrv> pQS(g_pScriptManager);
					if (pQS->Exists(pszStart))
					{
						lEventMask = pQS->Get(pszStart);
					}
				}
			}
			else
			{
				lEventMask = strtol(pszParams, NULL, 2);
			}
			g_pMalloc->Free(pszParams);
		}
		long lEventFlags = pInput->GetValidEvents();
		pInput->SetValidEvents(lEventFlags & lEventMask);
		pInvSrv->CapabilityControl(kDrkInvCapCycle, kDrkInvControlOff);
		pInvSrv->CapabilityControl(kDrkInvCapInvFrob, kDrkInvControlOff);

		if (m_iTiming == 0)
			m_iTiming = 10000;
	}
	else
	{
		pInput->SetValidEvents(0xFFFFFFFF);
		pInvSrv->CapabilityControl(kDrkInvCapCycle, kDrkInvControlOn);
		pInvSrv->CapabilityControl(kDrkInvCapInvFrob, kDrkInvControlOn);
	}
	return 0;
}


/****************
 * TrapCamAttach
 */
long cScr_CamAttach::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (bTurnOn)
	{
		if (GetObjectParamInt(m_iObjId, "nolens"))
		{
			SService<IDebugScrSrv> pDebugSrv(g_pScriptManager);
			char sz[32];
			sprintf(sz, "cam_attach %d", m_iObjId);
			pDebugSrv->Command(sz, "", "", "", "", "", "", "");
		}
		else
		{
			SService<ICameraSrv> pCam(g_pScriptManager);
			if (GetObjectParamInt(m_iObjId, "static"))
				pCam->StaticAttach(m_iObjId);
			else
				pCam->DynamicAttach(m_iObjId);
		}
	}
	else
	{
		SService<ICameraSrv> pCam(g_pScriptManager);
		if (GetObjectParamInt(m_iObjId, "forcereturn"))
			pCam->ForceCameraReturn();
		else
			pCam->CameraReturn(m_iObjId);
	}
	return 0;
}


/*******************
 * TrapFreezePlayer
 */
long cScr_FreezePlayer::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (m_iIsFrozen == 0)
	{
		SService<IDarkInvSrv> pDISrv(g_pScriptManager);
		pDISrv->AddSpeedControl("Freeze", 0.0, 0.0);
		m_iIsFrozen = 1;
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long cScr_FreezePlayer::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (m_iIsFrozen)
	{
		SService<IDarkInvSrv> pDISrv(g_pScriptManager);
		pDISrv->RemoveSpeedControl("Freeze");
		m_iIsFrozen = 0;
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long cScr_FreezePlayer::OnScale(float fValue, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IDarkInvSrv> pDISrv(g_pScriptManager);
	if (m_iIsFrozen)
		pDISrv->RemoveSpeedControl("Freeze");
	pDISrv->AddSpeedControl("Freeze", fValue, fValue);
	m_iIsFrozen = 1;
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*******************
 * TrapThreshold
 */
long __stdcall cScr_ThresholdRelay::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	m_iGlobalCounter.Init(GetObjectParamInt(m_iObjId, "initial"));
	return 0;
}

long cScr_ThresholdRelay::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	int iThreshold = GetObjectParamInt(m_iObjId, "threshold");
	if (GetObjectParamInt(m_iObjId, "global") || pMsg->from == 0)
	{
		if (++m_iGlobalCounter == iThreshold)
		{
			m_iGlobalCounter = 0;
			CDSend("TurnOn", m_iObjId);
			return 0;
		}
	}
	else
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		cMultiParm mpCounter;
		link lCounter;
		pLS->GetOne(lCounter, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, pMsg->from);
		if (!lCounter)
		{
			pLS->Create(lCounter, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, pMsg->from);
			if (!lCounter) // good candidate for throwing an exception here
				return 1;
			mpCounter = GetObjectParamString(m_iObjId, "initial");
			pLTS->LinkSetData(lCounter, NULL, mpCounter);
		}
		else
		{
			pLTS->LinkGetData(mpCounter, lCounter, NULL);
		}
		int iCounter = 0;
		if (mpCounter.psz)
		{
			iCounter = strtol(static_cast<const char*>(mpCounter), NULL, 10);
			g_pMalloc->Free(mpCounter.psz);
		}
		if (++iCounter == iThreshold)
		{
			CDSend("TurnOn", m_iObjId);
			mpCounter = "0";
			pLTS->LinkSetData(lCounter, NULL, mpCounter);
			return 0;
		}
		char psz[16];
		sprintf(psz, "%d", iCounter);
		mpCounter = psz;
		pLTS->LinkSetData(lCounter, NULL, mpCounter);
	}
	m_iFlags &= ~kTrapFlagOnce;
	return 0;
}

long cScr_ThresholdRelay::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	int iFloor = GetObjectParamInt(m_iObjId, "floor");
	if (GetObjectParamInt(m_iObjId, "global") || pMsg->from == 0)
	{
		if (m_iGlobalCounter > iFloor)
			--m_iGlobalCounter;
	}
	else
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		cMultiParm mpCounter;
		link lCounter;
		pLS->GetOne(lCounter, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, pMsg->from);
		if (!lCounter)
		{
			pLS->Create(lCounter, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, pMsg->from);
			if (!lCounter) // good candidate for throwing an exception here
				return 1;
			mpCounter = GetObjectParamString(m_iObjId, "floor");
			pLTS->LinkSetData(lCounter, NULL, mpCounter);
		}
		else
		{
			pLTS->LinkGetData(mpCounter, lCounter, NULL);
		}
		int iCounter = 0;
		if (mpCounter.psz)
		{
			iCounter = strtol(static_cast<const char*>(mpCounter), NULL, 10);
			g_pMalloc->Free(mpCounter.psz);
		}
		if (iCounter > iFloor)
		{
			char psz[16];
			sprintf(psz, "%d", iCounter - 1);
			mpCounter = psz;
			pLTS->LinkSetData(lCounter, NULL, mpCounter);
		}
	}
	m_iFlags &= ~kTrapFlagOnce;
	return 0;
}


/*******************
 * TrapRandomTimer
 */
// Relay to all CD links.
long __stdcall cScr_RandomTimer::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	const char* pszMsgName = static_cast<const char*>(pTimerMsg->name);
	if (stricmp(pszMsgName, "TurnOn") && 
		stricmp(pszMsgName, "TurnOff"))
		return 1;
	
	SService<ILockSrv> pLock(g_pScriptManager);
	if (pLock->IsLocked(m_iObjId))
		return 1;

	RelayMessage(pszMsgName);
	return 0;
}

// Delay for a random amount of time.
long __stdcall cScr_RandomTimer::OnTurnOnOff(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	if (!pLock->IsLocked(m_iObjId))
	{
		unsigned long lFlags = 0;
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
		{
			cMultiParm mpFlags;
			pPropSrv->Get(mpFlags, m_iObjId, "TrapFlags", NULL);
			lFlags = static_cast<int>(mpFlags);
		}
		const char* pszMsgName = pMsg->message;
		if ((lFlags & kTrapFlagNoOn) && !stricmp(pszMsgName, "TurnOn"))
			return 1;
		if ((lFlags & kTrapFlagNoOff) && !stricmp(pszMsgName, "TurnOff"))
			return 1;
		if (lFlags & kTrapFlagInvert)
		{
			if (!stricmp(pszMsgName, "TurnOn"))
				pszMsgName = "TurnOff";
			else
				pszMsgName = "TurnOn";
		}

		int iMinTime = 0, iMaxTime = 0;
		char* pszParamString = GetObjectParams(m_iObjId);
		if (pszParamString)
		{
			iMinTime = GetParamInt(pszParamString, "min_time");
			iMaxTime = GetParamInt(pszParamString, "max_time");
			g_pMalloc->Free(pszParamString);
		}

		unsigned long iInterval;
		if (iMinTime >= iMaxTime)
			iInterval = iMaxTime;
		else
		{
			SService<IDataSrv> pDS(g_pScriptManager);
			iInterval = pDS->RandInt(iMinTime, iMaxTime);
		}
		if (iInterval > 0)
			g_pScriptManager->SetTimedMessage2(m_iObjId, pszMsgName, iInterval, kSTM_OneShot, g_mpUndef);
		else
			RelayMessage(pszMsgName);
	}
	return 0;
}

void cScr_RandomTimer::RelayMessage(const char* pszMsgName)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "TrapFlags", NULL);
		if (static_cast<int>(mpFlags) & kTrapFlagOnce)
		{
			mpFlags = 1;
			pPropSrv->Set(m_iObjId, "Locked", mpFlags);
		}
	}
	CDSend(pszMsgName, m_iObjId);
}


/***********************
 * TrapRepeatingRelay
 */
// Send TurnOn to all CD links.
// Then decrement the param and, if zero, stop the timer.
long __stdcall cScr_RepeatingRelay::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	const char* pszMsgName = static_cast<const char*>(pTimerMsg->name);
	if (stricmp(pszMsgName, "Repeat"))
		return 1;
	
	CDSend("TurnOn", m_iObjId);
	if (--m_iRepeatCount == 0)
	{
		if (m_hTimerHandle != NULL)
		{
			g_pScriptManager->KillTimedMessage(m_hTimerHandle);
			m_hTimerHandle = NULL;
		}
	}
	
	return 0;
}

long __stdcall cScr_RepeatingRelay::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_hTimerHandle == NULL)
		return 1;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "TrapFlags", NULL);
		if (!(static_cast<int>(mpFlags) & kTrapFlagNoOff))
		{
			g_pScriptManager->KillTimedMessage(m_hTimerHandle);
			m_hTimerHandle = NULL;
		}
	}
	else
	{
		g_pScriptManager->KillTimedMessage(m_hTimerHandle);
		m_hTimerHandle = NULL;
	}
	return 0;
}

// Send the message, then set a timer.
// Params are "repeat" and "interval".
long __stdcall cScr_RepeatingRelay::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	if (pLock->IsLocked(m_iObjId))
		return 1;

	int iCount = 0;
	unsigned long iInterval = 0;
	char* pszParamString = GetObjectParams(m_iObjId);
	if (pszParamString)
	{
		iCount = GetParamInt(pszParamString, "repeat");
		iInterval = GetParamInt(pszParamString, "interval");
		g_pMalloc->Free(pszParamString);
	}

	if (iCount > 0)
	{
		if (m_hTimerHandle == NULL)
		{
			m_iRepeatCount = --iCount;
			CDSend("TurnOn", m_iObjId);
			if (iCount)
			{
				m_hTimerHandle = g_pScriptManager->SetTimedMessage2(m_iObjId, "Repeat", iInterval, kSTM_Periodic, g_mpUndef);
			}
		}
		else // just prolong the existing timer
			m_iRepeatCount += iCount;
	}

	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "TrapFlags", NULL);
		if (static_cast<int>(mpFlags) & kTrapFlagOnce)
		{
			mpFlags = 1;
			pPropSrv->Set(m_iObjId, "Locked", mpFlags);
		}
	}

	return 0;
}


/*********************
 * TrapRenderFlash
 */
int cScr_RenderFlash::Relink(ILinkSrv* pLS, ILinkToolsSrv* pLTS, long flavor, int source, int dest)
{
	link l;
	pLS->GetOne(l, flavor, source, 0);
	int ret = 0;
	if (l)
	{
		sLink sl;
		pLTS->LinkGet(l, &sl);
		ret = sl.dest;
		pLS->Destroy(l);
	}
	link nl;
	// RenderFlash links don't have any data, so no worries.
	pLS->Create(nl, flavor, source, dest);
	return ret;
}

long cScr_RenderFlash::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (!bTurnOn)
		return 1;
	
	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	object oPlayer;
	pObjSrv->Named(oPlayer, "Player");
	// A bit of trickery here.  The renderflash is linked from the archetype of "Player",
	// not the concrete object. 
	SInterface<ITraitManager> pTraits(g_pScriptManager);
	int iAvatar = pTraits->GetArchetype(oPlayer);
	if (iAvatar != 0 && oPlayer != 0)
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		// The render flash that is linked from "Garrett" will be shown.
		long flavor = pLTS->LinkKindNamed("RenderFlash");
		int iObject = m_iObjId;
		link l = 0;
		pLS->GetOne(l, flavor, iObject, 0);
		while (l == 0)
		{
			iObject = pTraits->GetArchetype(iObject);
			true_bool valid_p;
			if (!*(pObjSrv->Exists(valid_p, iObject)))
				break;
			pLS->GetOne(l, flavor, iObject, 0);
		}
		if (l != 0)
		{
			sLink sl;
			pLTS->LinkGet(l, &sl);
			int oldflash = Relink(pLS, pLTS, flavor, iAvatar, sl.dest);
			SService<ICameraSrv> pCam(g_pScriptManager);
			SService<ISoundScrSrv> pSound(g_pScriptManager);
			true_bool ret;
			int iSchema = GetOneLinkDest("SoundDescription", m_iObjId, NULL);
			if (iSchema)
				pSound->PlaySchemaAmbient(ret, oPlayer, iSchema, kSoundNetwork0);
			pCam->StaticAttach(oPlayer);
			pCam->CameraReturn(oPlayer);
			Relink(pLS, pLTS, flavor, iAvatar, oldflash);
		}
	}
	return 0;
}


/*******************
 * TrapDisarmPlayer
 */
int cScr_DisarmPlayer::DisarmIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	IObjectSrv* pOS = reinterpret_cast<IObjectSrv*>(pData);
	sLink sl;
	pLQ->Link(&sl);
	int iObjId = static_cast<cScr_DisarmPlayer*>(pScript)->m_iObjId;
	bool bPotions = true;
	char* pszParam = GetObjectParamString(iObjId, "nopotions");
	if (pszParam)
	{
		bPotions = strtol(pszParam, NULL, 10) == 0;
		g_pMalloc->Free(pszParam);
	}
	true_bool child_p;
	if (*(pOS->InheritsFrom(child_p, sl.dest, -235 )) ||
		*(pOS->InheritsFrom(child_p, sl.dest, -59  )) ||
		*(pOS->InheritsFrom(child_p, sl.dest, -140 )) ||
		*(pOS->InheritsFrom(child_p, sl.dest, -1953)) ||
		(bPotions && *(pOS->InheritsFrom(child_p, sl.dest, -2546))) )
	{
		SService<IContainSrv> pContainSrv(g_pScriptManager);
		pContainSrv->Remove(sl.dest, 0);
		pContainSrv->Add(sl.dest, iObjId, 0, 1);
	}
	return 1;
}

long cScr_DisarmPlayer::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	object oPlayer;
	pObjSrv->Named(oPlayer, "Player");
	if (bTurnOn)
	{
		IterateLinks("Contains", oPlayer, 0, DisarmIterFunc, this, reinterpret_cast<void*>(pObjSrv.get()));
	}
	else
	{
		SService<IContainSrv> pContainSrv(g_pScriptManager);
		pContainSrv->MoveAllContents(m_iObjId, oPlayer, 1);
	}
	return 0;
}


/*****************
 * NewCrystal
 */
long __stdcall cScr_NewCrystal::OnContained(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sContainedScrMsg* pContainMsg = static_cast<sContainedScrMsg*>(pMsg);
	if (pContainMsg->event != kContainAdd && pContainMsg->event != kContainCombine)
		return 1;

	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	object oAvatar;
	pObjSrv->Named(oAvatar, "Avatar");
	true_bool bIsAvatar;
	if (*(pObjSrv->InheritsFrom(bIsAvatar, pContainMsg->container, oAvatar)))
	{
		return TransmogrifyCrystal(pObjSrv, pContainMsg->container);
	}
	return 1;
}

long cScr_NewCrystal::TransmogrifyCrystal(IObjectSrv* pObjSrv, int iDest)
{
	sLink slArrow;
	int iObject = m_iObjId;
	SInterface<ITraitManager> pTraits(g_pScriptManager);
	while (!FindLinkByData("ScriptParams", iObject, 0, &slArrow, "Crystal", -1))
	{
		iObject = pTraits->GetArchetype(iObject);
		true_bool valid_p;
		if (!*(pObjSrv->Exists(valid_p, iObject)))
			return 1;
	}
	object oArrow;
	pObjSrv->Create(oArrow, slArrow.dest);
	if (oArrow == 0)
		return 1;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "StackCount"))
	{
		cMultiParm mpStack;
		pPropSrv->Get(mpStack, m_iObjId, "StackCount", NULL);
		pPropSrv->Set(oArrow, "StackCount", mpStack);
	}
	SService<IContainSrv> pContainSrv(g_pScriptManager);
	pContainSrv->Add(oArrow, iDest, 0, 1);
	pObjSrv->Destroy(m_iObjId);
	return 0;
}


/*******************
 * DoorFrobInert
 */
// Add or Remove the metaproperty based on the current door state.
long __stdcall cScr_DoorFrobInert::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		m_iDoorIsDead.Init(0);
		SService<IDoorSrv> pDoor(g_pScriptManager);
		int iState = pDoor->GetDoorState(m_iObjId);
		if (m_iDoorIsDead)
			iState = kDoorStateOpen;
		IterateLinksByData("ScriptParams", m_iObjId, 0, "FrobInert", -1, FrobInertIterFunc, this, reinterpret_cast<void*>(iState));
	}	
	return 1;
}

// Door is no more. Remove the metaproperty.
// Of course, if we get this but the door hasn't been opened,
// then there will likely be an ugly black rectangle where the door was.
// We make the objects frobable anyway.
long __stdcall cScr_DoorFrobInert::OnDestroy(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	//SService<IDoorSrv> pDS(g_pScriptManager);
	//int iState = pDS->GetDoorState(m_iObjId);
	//if (iState == kDoorStateClosed)
		IterateLinksByData("ScriptParams", m_iObjId, 0, "FrobInert", -1, FrobInertIterFunc, this, reinterpret_cast<void*>(kDoorStateOpen));
	if (m_bSim)
		m_iDoorIsDead = 1;
	return 0;
}

// Door is dead. We may or may not want to terminally remove the metaproperty.
// Check the "breakable" parameter to see if we should treat the door
// as permanently open.
long __stdcall cScr_DoorFrobInert::OnSlain(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (GetObjectParamInt(m_iObjId, "breakable"))
	{
		m_iDoorIsDead = 1;
		return 0;
	}
	return 1;
}

// Remove the metaproperty.
long __stdcall cScr_DoorFrobInert::OnDoorOpening(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	IterateLinksByData("ScriptParams", m_iObjId, 0, "FrobInert", -1, FrobInertIterFunc, this, reinterpret_cast<void*>(kDoorStateOpening));
	return 0;
}

// Add the metaproperty (if it doesn't already exist).
long __stdcall cScr_DoorFrobInert::OnDoorClose(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!m_iDoorIsDead)
		IterateLinksByData("ScriptParams", m_iObjId, 0, "FrobInert", -1, FrobInertIterFunc, this, reinterpret_cast<void*>(kDoorStateClosed));
	return 0;
}

int cScr_DoorFrobInert::FrobInertIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	int iState = reinterpret_cast<int>(pData);
	sLink sl;
	pLQ->Link(&sl);
	if (iState == kDoorStateClosed)
		AddSingleMetaProperty("FrobInert", sl.dest);
	else
		RemoveSingleMetaProperty("FrobInert", sl.dest);
	return 1;
}


/********************
 * UnlinkOnContained
 */
// Remove links from a specified object.
// Only happens once to save time.
long __stdcall cScr_ContainedUnlink::OnContained(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_iAlreadyUnlinked != 0)
		return 0;
	if (static_cast<sContainedScrMsg*>(pMsg)->event != kContainAdd && 
		static_cast<sContainedScrMsg*>(pMsg)->event != kContainCombine)
		return 1;

	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	object oAvatar;
	pObjSrv->Named(oAvatar, "Avatar");
	true_bool bIsAvatar;
	if (*(pObjSrv->InheritsFrom(bIsAvatar, static_cast<sContainedScrMsg*>(pMsg)->container, oAvatar)))
	{
		IterateLinksByData("ScriptParams", m_iObjId, 0, "Unlink", -1, ScriptParamsIterFunc, this, NULL);
		m_iAlreadyUnlinked = 1;
		return 0;
	}
	return 1;
}

int cScr_ContainedUnlink::ScriptParamsIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	sLink sl;
	pLQ->Link(&sl);
	IterateLinks(NULL, sl.dest, sl.source, DestroyLinksIterFunc, pScript, NULL);
	/* Still haven't worked out how to use DestroyMany
	char szFrom[12];
	char szTo[12];
	sprintf(szFrom, "%d", sl.dest);
	sprintf(szTo, "%d", sl.source);
	pLS->DestroyMany(0, szFrom, szTo);
	*/
	pLS->Destroy(pLQ->ID());
	return 1;
}

int cScr_ContainedUnlink::DestroyLinksIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	pLS->Destroy(pLQ->ID());
	return 1;
}


/******************
 * TimedGrenade
 */
long __stdcall cScr_TimedGrenade::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	m_iSafety.Init();
	m_hFuseTimer.Init();

	/* Bad idea. Use a receptron link to send to script
	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	SService<IActReactSrv> pARSrv(g_pScriptManager);
	object oStim;
	pObjSrv->Named(oStim, "FireStim");
	pARSrv->SubscribeToStimulus(m_iObjId, oStim);
	pObjSrv->Named(oStim, "BashStim");
	pARSrv->SubscribeToStimulus(m_iObjId, oStim);
	pObjSrv->Named(oStim, "PokeStim");
	pARSrv->SubscribeToStimulus(m_iObjId, oStim);
	*/
	if (m_hFuseTimer)
	{
		RegisterMessageHandler("BashStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
		RegisterMessageHandler("PokeStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
	}
	return 0;
}

long __stdcall cScr_TimedGrenade::OnEndScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	/*
	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	SService<IActReactSrv> pARSrv(g_pScriptManager);
	object oStim;
	pObjSrv->Named(oStim, "FireStim");
	pARSrv->UnsubscribeToStimulus(m_iObjId, oStim);
	pObjSrv->Named(oStim, "BashStim");
	pARSrv->UnsubscribeToStimulus(m_iObjId, oStim);
	pObjSrv->Named(oStim, "PokeStim");
	pARSrv->UnsubscribeToStimulus(m_iObjId, oStim);
	*/
	return 1;
}

long __stdcall cScr_TimedGrenade::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	const char* pszMsgName = static_cast<const char*>(pTimerMsg->name);

	if (!stricmp(pszMsgName, "StackHack"))
	{
		RemoveObjectParam(m_iObjId, "launch_active");
		return 0;
	}
	if (!stricmp(pszMsgName, "GrenadeActive"))
	{
		GrenadeActive(true);
		return 0;
	}
	if (stricmp(pszMsgName, "TimedGrenade"))
		return 1;

	if (m_iSafety)
	{
		m_iSafety = 0;
		m_hFuseTimer = g_pScriptManager->SetTimedMessage2(m_iObjId, "TimedGrenade", 1600, kSTM_OneShot, g_mpUndef);
	}
	else
	{
		m_hFuseTimer = NULL;
		SService<IDamageSrv> pDS(g_pScriptManager);
		pDS->Slay(m_iObjId, 0);
	}
	return 0;
}

long __stdcall cScr_TimedGrenade::OnStimulus(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	// Intensity is handled by the receptron link.
	//if (static_cast<sStimMsg*>(pMsg)->intensity < 5.0)
	//	return 1;
	SService<IDamageSrv> pDS(g_pScriptManager);
	pDS->Slay(m_iObjId, static_cast<sStimMsg*>(pMsg)->source);
	return 0;
}

long __stdcall cScr_TimedGrenade::OnSlain(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_hFuseTimer)
	{
		g_pScriptManager->KillTimedMessage(m_hFuseTimer);
		m_hFuseTimer = NULL;
	}
	SService<ISoundScrSrv> pSoundSrv(g_pScriptManager);
	true_bool bRet;
	pSoundSrv->HaltSchema(bRet, m_iObjId, "", 0);
	
	GrenadeFlash();

	return 0;
}

long __stdcall cScr_TimedGrenade::OnFrobInv(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "StackCount"))
	{
		cMultiParm mpStack;
		pPropSrv->Get(mpStack, m_iObjId, "StackCount", NULL);
		if (static_cast<int>(mpStack) > 1)
		{
			// This is gonna get ugly.
			char* pszParams = GetObjectParams(m_iObjId);
			char* pszNew;
			if (pszParams)
			{
				pszNew = new char[strlen(pszParams) + 17];
				sprintf(pszNew, "%s;%s", "launch_active=1", pszParams);
			}
			else
			{
				pszNew = new char[16];
				strcpy(pszNew, "launch_active=1");
			}
			SetObjectParams(m_iObjId, pszNew);
			delete[] pszNew;
			g_pScriptManager->SetTimedMessage2(m_iObjId, "StackHack", 1, kSTM_OneShot, g_mpUndef);
		}
		else
		{
			GrenadeArmed();
		}
	}
	else
	{
		GrenadeArmed();
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long __stdcall cScr_TimedGrenade::OnFrobWorld(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	if (m_hFuseTimer)
	{
		g_pScriptManager->KillTimedMessage(m_hFuseTimer);
		m_hFuseTimer = NULL;
		if (m_iSafety)
		{
			m_iSafety = 0;
			GrenadeActive(false);
		}
		else
		{
			SService<IDamageSrv> pDS(g_pScriptManager);
			pDS->Slay(m_iObjId, static_cast<sStimMsg*>(pMsg)->source);
		}
		return 0;
	}
	// Just in case
	return 1;
}

long __stdcall cScr_TimedGrenade::OnContained(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	// Activate the schema after being removed from the inventory.
	if (static_cast<sContainedScrMsg*>(pMsg)->event == kContainRemove && m_hFuseTimer)
	{
		GrenadeActive(true);
		return 0;
	}
	else if (static_cast<sContainedScrMsg*>(pMsg)->event == kContainAdd
		  || static_cast<sContainedScrMsg*>(pMsg)->event == kContainCombine)
	{
		RemoveObjectParam(m_iObjId, "launch_active");
	}
	return 1;
}

long __stdcall cScr_TimedGrenade::OnCreate(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	// Dynamically created objects don't get a Sim message. Use the message time instead.
	if (pMsg->time > 0 && GetObjectParamInt(m_iObjId, "launch_active"))
	{
		GrenadeArmed();
		g_pScriptManager->SetTimedMessage2(m_iObjId, "GrenadeActive", 1, kSTM_OneShot, g_mpUndef);
		RemoveObjectParam(m_iObjId, "launch_active");
		return 0;
	}
	return 1;
}

long __stdcall cScr_TimedGrenade::OnSchemaDone(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_hFuseTimer) // We're active, so let's play the active loop.
	{
		SService<ISoundScrSrv> pSoundSrv(g_pScriptManager);
		true_bool bRet;
		pSoundSrv->PlayEnvSchema(bRet, m_iObjId, "Event ActiveLoop", m_iObjId, 0, kEnvSoundOnObj, kSoundNetwork0);
		return 0;
	}
	// If not active... who cares?
	return 1;
}

void cScr_TimedGrenade::GrenadeArmed(void)
{
	if (m_hFuseTimer)
	{
		// How in the hell did you get here? Well, let's clean this up.
		g_pScriptManager->KillTimedMessage(m_hFuseTimer);
		m_hFuseTimer = NULL;
	}
	int iFuseTime = 5000;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "ScriptTiming"))
	{
		cMultiParm mpTiming;
		pPropSrv->Get(mpTiming, m_iObjId, "ScriptTiming", NULL);
		iFuseTime = mpTiming;
	}
	if (iFuseTime > 2000)
	{
		m_iSafety = 1;
		iFuseTime -= 1600;
	}
	else // Uh-oh...
	{
		m_iSafety = 0;
	}
	m_hFuseTimer = g_pScriptManager->SetTimedMessage2(m_iObjId, "TimedGrenade", iFuseTime, kSTM_OneShot, g_mpUndef);
}

void cScr_TimedGrenade::GrenadeFlash(void)
{
	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	SInterface<ITraitManager> pTraits(g_pScriptManager);
	long flavor = pLTS->LinkKindNamed("RenderFlash");
	// TriggerWorldFlash ignores the RenderFlash link on the concrete object
	// ... the bastard
	int iObject = pTraits->GetArchetype(m_iObjId);
	link lFlash;
	pLS->GetOne(lFlash, flavor, iObject, 0);
	while (lFlash == 0)
	{
		iObject = pTraits->GetArchetype(iObject);
		true_bool valid_p;
		if (!*(pObjSrv->Exists(valid_p, iObject)))
			break;
		pLS->GetOne(lFlash, flavor, iObject, 0);
	}
	if (lFlash != 0)
	{
		SService<IDarkPowerupsSrv> pDPS(g_pScriptManager);
		pDPS->TriggerWorldFlash(m_iObjId);
	}
}

void cScr_TimedGrenade::GrenadeActive(bool bActivate)
{

	SService<ISoundScrSrv> pSoundSrv(g_pScriptManager);
	true_bool bRet;
	if (bActivate)
	{
		//pSoundSrv->HaltSchema(bRet, m_iObjId, "", 0);
		pSoundSrv->PlayEnvSchema(bRet, m_iObjId, "Event Activate", m_iObjId, 0, kEnvSoundOnObj, kSoundNetwork0);
		if (!bRet)
		{
			pSoundSrv->PlayEnvSchema(bRet, m_iObjId, "Event ActiveLoop", m_iObjId, 0, kEnvSoundOnObj, kSoundNetwork0);
		}
	}
	else
	{
		pSoundSrv->HaltSchema(bRet, m_iObjId, "", 0);
		pSoundSrv->PlayEnvSchema(bRet, m_iObjId, "Event Deactivate", m_iObjId, 0, kEnvSoundOnObj, kSoundNetwork0);
	}

	SService<IActReactSrv> pARSrv(g_pScriptManager);
	cMultiParm& mpNull = const_cast<cMultiParm&>(g_mpUndef);
	cMultiParm mpType = kTweqTypeModels;
	cMultiParm mpAction = (bActivate) ? kTweqDoActivate : kTweqDoHalt;
	pARSrv->React(pARSrv->GetReactionNamed("tweq_control"), 1.0, m_iObjId, 0,
			  mpType, mpAction, mpNull, mpNull, mpNull, mpNull, mpNull, mpNull);
	if (!bActivate)
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		if (pPropSrv->Possessed(m_iObjId, "CfgTweqModels"))
		{
			cMultiParm mpModel;
			pPropSrv->Get(mpModel, m_iObjId, "CfgTweqModels", "Model 5");
			if (mpModel.type == kMT_String && mpModel.psz != NULL)
			{
				pPropSrv->Set(m_iObjId, "ModelName", mpModel);
				g_pMalloc->Free(mpModel.psz);
			}
		}
	}

	if (bActivate)
	{
		RegisterMessageHandler("BashStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
		RegisterMessageHandler("PokeStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
	}
	else
	{
		UnregisterMessageHandler("BashStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
		UnregisterMessageHandler("PokeStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
	}
}


/****************
 * ZeroGravRoom
 *
 * This script was reverse-engineered from System Shock 2.
 * (In particular, I never would've known to set the base friction.)
 */
long __stdcall cScr_ZeroGravRoom::OnEnter(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sRoomMsg* pRMsg = static_cast<sRoomMsg*>(pMsg);
	SService<IPropertySrv> pPSrv(g_pScriptManager);
	pPSrv->Set(pRMsg->MoveObjId, "PhysAttr", "Gravity %", 0.0);
	pPSrv->Set(pRMsg->MoveObjId, "PhysAttr", "Base Friction", 1.0);

	return 0;
}

long __stdcall cScr_ZeroGravRoom::OnExit(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sRoomMsg* pRMsg = static_cast<sRoomMsg*>(pMsg);
	SService<IPropertySrv> pPSrv(g_pScriptManager);
	float fGrav = 100.0;
	if (pPSrv->Possessed(pRMsg->ToObjId, "RoomGrav"))
	{
		cMultiParm mpGrav;
		pPSrv->Get(mpGrav, pRMsg->ToObjId, "RoomGrav", NULL);
		fGrav = static_cast<int>(mpGrav);
	}
	pPSrv->Set(pRMsg->MoveObjId, "PhysAttr", "Gravity %", fGrav);

	return 0;
}


/***************
 * TrapFadeOut
 */
long cScr_FadeOut::OnScale(float fValue, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IDarkGameSrv> pDG(g_pScriptManager);
	pDG->FadeToBlack(fValue);
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long cScr_FadeOut::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IDarkGameSrv> pDG(g_pScriptManager);
	pDG->FadeToBlack(-1);
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*****************
 * KnockOnDoor
 */
long __stdcall cScr_KnockOnDoor::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLockSrv(g_pScriptManager);
	SService<IDoorSrv> pDoorSrv(g_pScriptManager);
	if (pLockSrv->IsLocked(m_iObjId) && pDoorSrv->GetDoorState(m_iObjId) == kDoorStateClosed)
	{
		SService<IAIScrSrv> pAISrv(g_pScriptManager);
		IterateLinks("Owns", 0, m_iObjId, LinkIterFunc, this, reinterpret_cast<void*>(pAISrv.get()));
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

int cScr_KnockOnDoor::LinkIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	IAIScrSrv* pAISrv = reinterpret_cast<IAIScrSrv*>(pData);
	sLink sl;
	pLQ->Link(&sl);
	pAISrv->Signal(sl.source, "knock_knock");
	return 1;
}


/*****************
 * LockHelper
 */
long __stdcall cScr_LockHelper::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLockSrv(g_pScriptManager);
	if (pLockSrv->IsLocked(m_iObjId))
		IterateLinksByData("ScriptParams", m_iObjId, 0, "FrobLock", -1, LinkIterFunc, this, reinterpret_cast<void*>(1));
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long __stdcall cScr_LockHelper::OnMessage(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	// ILinkSrv::BroadcastOnAllLinks will do this, but it uses a case-sensitive compare.
	// I don't want to be that strict.
	SService<ILockSrv> pLockSrv(g_pScriptManager);
	IterateLinksByData("ScriptParams", m_iObjId, 0, pMsg->message, -1, LinkIterFunc,
			    this, reinterpret_cast<void*>(pLockSrv->IsLocked(m_iObjId)));
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

int cScr_LockHelper::LinkIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	int iLocked = reinterpret_cast<int>(pData);
	sLink sl;
	pLQ->Link(&sl);
	SimplePost(sl.source, sl.dest, "TurnOn", iLocked);
	return 1;
}


/*****************
 * KeyHelper
 */
long __stdcall cScr_KeyHelper::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SimpleSend(m_iObjId, static_cast<sFrobMsg*>(pMsg)->DstObjId, "UseKey");
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*****************
 * PickHelper
 */
long __stdcall cScr_PickHelper::OnFrobInvBegin(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SimpleSend(m_iObjId, static_cast<sFrobMsg*>(pMsg)->DstObjId, "PickBegin");
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long __stdcall cScr_PickHelper::OnFrobInvEnd(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SimpleSend(m_iObjId, static_cast<sFrobMsg*>(pMsg)->DstObjId, "PickEnd");
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/******************
 * TrapSiblingRelay
 */
long __stdcall cScr_SiblingRelay::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	int iInit = 0;
	int iLowestSibling = FindFirstObject(m_iObjId);
	m_iBaseObjectId.Init(0);
	m_iCurrentState.Init(GetObjectParamInt(m_iObjId, "start_on"));
	m_iSwallowTurnOn.Init(0);
	m_iSwallowTurnOff.Init(0);
	if (m_iBaseObjectId == 0)
	{
		m_iBaseObjectId = iLowestSibling;
	}
	else if (m_iBaseObjectId != iLowestSibling)
	{
		// Joy. The base object ID has changed while we weren't looking.
		if (iLowestSibling == m_iObjId)
		{
			sScrDatumTag oldLock = { m_iBaseObjectId,"cScr_SiblingRelay","m_giSiblingLock"};
			if (g_pScriptManager->IsScriptDataSet(&oldLock))
			{
				sMultiParm param;
				param.type = kMT_Int;
				g_pScriptManager->ClearScriptData(&oldLock, &param);
				iInit = param.i;
			}
		}
		m_iBaseObjectId = iLowestSibling;
	}
	m_giSiblingLock.SetObj(iLowestSibling);
	if (iLowestSibling == m_iObjId)
		m_giSiblingLock.Init(iInit);

	return 0;
}

long __stdcall cScr_SiblingRelay::OnDestroy(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bSim && m_iBaseObjectId == m_iObjId)
	{
		int iNewBase = FindFirstObject(0);
		if (iNewBase != 0)
		{
			int iSiblingLock = m_giSiblingLock;
			m_giSiblingLock.SetObj(iNewBase);
			m_giSiblingLock = iSiblingLock;
			SService<IPropertySrv> pPropSrv(g_pScriptManager);
			if (pPropSrv->Possessed(m_iObjId, "Locked"))
				pPropSrv->CopyFrom(iNewBase, "Locked", m_iObjId);
			else
				pPropSrv->Remove(iNewBase, "Locked");
			if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
				pPropSrv->CopyFrom(iNewBase, "TrapFlags", m_iObjId);
			else
				pPropSrv->Remove(iNewBase, "TrapFlags");
			if (pPropSrv->Possessed(m_iObjId, "ScriptTiming"))
				pPropSrv->CopyFrom(iNewBase, "ScriptTiming", m_iObjId);
			else
				pPropSrv->Remove(iNewBase, "ScriptTiming");
			PropogateMessage("NewBaseObject", iNewBase);
		}
	}
	// Let's go ahead and re-link outside of sim-time.
	// Makes maintenance of networks in DromEd easier.
	int iLeft = 0, iRight = 0;
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed("ScriptParams");
	ILinkQuery* pLQ = pRel->Query(0, m_iObjId);
	for (; !pLQ->Done(); pLQ->Next())
	{
		if (!stricmp(reinterpret_cast<char*>(pLQ->Data()), "Sibling"))
		{
			sLink sl;
			pLQ->Link(&sl);
			iLeft = sl.source;
			break;
		}
	}
	pLQ->Release();
	pLQ = pRel->Query(m_iObjId, 0);
	for (; !pLQ->Done(); pLQ->Next())
	{
		if (!stricmp(reinterpret_cast<char*>(pLQ->Data()), "Sibling"))
		{
			sLink sl;
			pLQ->Link(&sl);
			iRight = sl.dest;
			break;
		}
	}
	pLQ->Release();
	if (iLeft && iRight)
		pRel->AddFull(iLeft, iRight, "Sibling");

	return 0;
}

long __stdcall cScr_SiblingRelay::OnNewBaseObject(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (pMsg->data.type == kMT_Int && pMsg->data.i != 0)
	{
		m_iBaseObjectId = pMsg->data.i;
		m_giSiblingLock.SetObj(pMsg->data.i);
	}
	return 0;
}

long __stdcall cScr_SiblingRelay::OnNowLocked(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bSim && m_iBaseObjectId != m_iObjId)
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		cMultiParm mpLock = 1;
		pPropSrv->Set(static_cast<int>(m_iBaseObjectId), "Locked", mpLock);
	}
	return 0;
}

long __stdcall cScr_SiblingRelay::OnNowUnlocked(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bSim && m_iBaseObjectId != m_iObjId)
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		cMultiParm mpLock = 0;
		pPropSrv->Set(static_cast<int>(m_iBaseObjectId), "Locked", mpLock);
	}
	return 0;
}

void cScr_SiblingRelay::InitTrapVars()
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	int iObj = m_iBaseObjectId;
	if (pPropSrv->Possessed(iObj, "ScriptTiming"))
	{
		cMultiParm mpTiming = 0;
		pPropSrv->Get(mpTiming, iObj, "ScriptTiming", NULL);
		m_iTiming = mpTiming;
	}
	if (pPropSrv->Possessed(iObj, "TrapFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, iObj, "TrapFlags", NULL);
		m_iFlags = static_cast<int>(mpTrapFlags);
	}
}

// As well as reading from the base object, we always want to allow propogation msgs.
bool cScr_SiblingRelay::IsLocked(sScrMsg* pMsg)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	return m_giSiblingLock == 0 && pLock->IsLocked(static_cast<int>(m_iBaseObjectId));
}

void cScr_SiblingRelay::SetLock(bool bLock)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	cMultiParm mpLock = static_cast<int>(bLock);
	pPropSrv->Set(static_cast<int>(m_iBaseObjectId), "Locked", mpLock);
}

long cScr_SiblingRelay::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (m_iSwallowTurnOn > 0)
	{
		--m_iSwallowTurnOn;
	}
	else
	{
		CDSend("TurnOn", m_iObjId);
		if (m_giSiblingLock == 0)
		{
			m_giSiblingLock = 1;
			const char* pszOn = "TurnOn";
			const char* pszOff = "TurnOff";
			if (m_iFlags & kTrapFlagInvert)
			{
				pszOn = "TurnOff";
				pszOff = "TurnOn";
			}
			char* pszParam = GetObjectParamString(m_iObjId, "exclusive");
			if (pszParam)
			{
				if (!stricmp(pszParam, "on") || !stricmp(pszParam, "all"))
					PropogateMessage(pszOff, 1);
				else
					PropogateMessage(pszOn, 1);
				g_pMalloc->Free(pszParam);
			}
			else
				PropogateMessage(pszOn, 1);
			m_giSiblingLock = 0;
			
			// Must be from Timer
			if (pMsg->from == 0)
				ToggleLevers(1);
		}
		else // m_giSiblingLock
		{
			if (m_iCurrentState != 1)
				ToggleLevers(1);
			m_iTiming = 0;
			m_iFlags = 0;
		}
		m_iCurrentState = 1;
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long cScr_SiblingRelay::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (m_iSwallowTurnOff > 0)
	{
		--m_iSwallowTurnOff;
	}
	else
	{
		CDSend("TurnOff", m_iObjId);
		if (m_giSiblingLock == 0)
		{
			m_giSiblingLock = 1;
			const char* pszOn = "TurnOn";
			const char* pszOff = "TurnOff";
			if (m_iFlags & kTrapFlagInvert)
			{
				pszOn = "TurnOff";
				pszOff = "TurnOn";
			}
			char* pszParam = GetObjectParamString(m_iObjId, "exclusive");
			if (pszParam)
			{
				if (!stricmp(pszParam, "off") || !stricmp(pszParam, "all"))
					PropogateMessage(pszOn, 1);
				else
					PropogateMessage(pszOff, 1);
				g_pMalloc->Free(pszParam);
			}
			else
				PropogateMessage(pszOff, 1);
			m_giSiblingLock = 0;
			
			// Must be from Timer
			if (pMsg->from == 0)
				ToggleLevers(0);
		}
		else // m_giSiblingLock
		{
			if (m_iCurrentState != 0)
				ToggleLevers(0);
			m_iTiming = 0;
			m_iFlags = 0;
		}
		m_iCurrentState = 0;
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

struct SiblingRelayToggle_Struct
{
	int iState;
	const char* pszMsg;
};

int cScr_SiblingRelay::ToggleIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	cScr_SiblingRelay* pRelay = static_cast<cScr_SiblingRelay*>(pScript);
	SiblingRelayToggle_Struct* pToggleData = reinterpret_cast<SiblingRelayToggle_Struct*>(pData);
	if (pToggleData->iState > 0)
		++pRelay->m_iSwallowTurnOn;
	else if (pToggleData->iState == 0)
		++pRelay->m_iSwallowTurnOff;
	sLink sl;
	pLQ->Link(&sl);
	// Oh hell, let's be greedy.
	SimpleSend(pRelay->m_iObjId, sl.dest, pToggleData->pszMsg);
	return 1;
}

// This pisses me off. How friggen' hard is it to synchronize the state of
// a lever without creating a message loop.
void cScr_SiblingRelay::ToggleLevers(int iState)
{
	SiblingRelayToggle_Struct sToggleData;
	int iLeverState = iState ^ (GetObjectParamInt(m_iObjId, "invert_levers") != 0);
	iLeverState ^= ((m_iFlags & kTrapFlagInvert) == kTrapFlagInvert);
	sToggleData.pszMsg = (iLeverState) ? "TurnOn" : "TurnOff";
	if (GetObjectParamInt(m_iObjId, "no_swallow"))
		sToggleData.iState = -1;
	else
		sToggleData.iState = iState;

	IterateLinks("Owns", m_iObjId, 0, ToggleIterFunc, this, reinterpret_cast<void*>(&sToggleData));
}

// Hell, at least this part worked without incident
void cScr_SiblingRelay::PropogateMessage(const char* pszMsg, int param)
{
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed("ScriptParams");
	ILinkQuery* pLQ = pRel->Query(m_iObjId, 0);
	while (!pLQ->Done())
	{
		if (!stricmp(reinterpret_cast<char*>(pLQ->Data()), "Sibling"))
		{
			sLink slInfo;
			pLQ->Link(&slInfo);
			if (slInfo.dest == m_iObjId)
				break;
			// I think I'll have the originating object send the message.
			// Rather than going peer-to-peer.
			SimpleSend(m_iObjId, slInfo.dest, pszMsg, param);
			pLQ->Release();
			pLQ = pRel->Query(slInfo.dest, 0);
			continue;
		}
		pLQ->Next();
	}
	pLQ->Release();
}

// Traverse the sibling links looking for the lowest object ID.
int cScr_SiblingRelay::FindFirstObject(int iBase)
{
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed("ScriptParams");
	ILinkQuery* pLQ = pRel->Query(m_iObjId, 0);
	while (!pLQ->Done())
	{
		if (!stricmp(reinterpret_cast<char*>(pLQ->Data()), "Sibling"))
		{
			sLink slInfo;
			pLQ->Link(&slInfo);
			if (slInfo.dest == m_iObjId)
				break;
			if (iBase > slInfo.dest || iBase == 0)
				iBase = slInfo.dest;
			pLQ->Release();
			pLQ = pRel->Query(slInfo.dest, 0);
			continue;
		}
		pLQ->Next();
	}
	pLQ->Release();
	return iBase;
}


/******************
 * TrigContact
 */
long __stdcall cScr_ContactTrigger::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPhysSrv> pPhysSrv(g_pScriptManager);
	pPhysSrv->SubscribeMsg(cScript::m_iObjId, kPhysContact);
	return 0;
}

long __stdcall cScr_ContactTrigger::OnEndScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPhysSrv> pPhysSrv(g_pScriptManager);
	pPhysSrv->UnsubscribeMsg(cScript::m_iObjId, kPhysContact);
	return 0;
}

struct ContactTrigger_Struct
{
	bool bOn;
	int iObject;
};

long __stdcall cScr_ContactTrigger::OnPhysContactCreate(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!IsLocked(pMsg))
	{
		ContactTrigger_Struct sContactInfo;
		if (0 == (sContactInfo.iObject = static_cast<sPhysMsg*>(pMsg)->contactObj))
			return 1;
		sContactInfo.bOn = true;
		InitTrigVars();
		if (0 == IterateLinksByData("ScriptParams", cScript::m_iObjId, sContactInfo.iObject, "Contact", -1, FindContactObject, this, reinterpret_cast<void*>(&sContactInfo)))
		{
			IterateLinksByData("ScriptParams", cScript::m_iObjId, 0, "Contact", -1, FindContactArchetype, this, reinterpret_cast<void*>(&sContactInfo));
		}
		return 0;
	}
	return 1;
}

long __stdcall cScr_ContactTrigger::OnPhysContactDestroy(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!IsLocked(pMsg))
	{
		ContactTrigger_Struct sContactInfo;
		if (0 == (sContactInfo.iObject = static_cast<sPhysMsg*>(pMsg)->contactObj))
			return 1;
		sContactInfo.bOn = false;
		InitTrigVars();
		if (0 == IterateLinksByData("ScriptParams", cScript::m_iObjId, sContactInfo.iObject, "Contact", -1, FindContactObject, this, reinterpret_cast<void*>(&sContactInfo)))
		{
			IterateLinksByData("ScriptParams", cScript::m_iObjId, 0, "Contact", -1, FindContactArchetype, this, reinterpret_cast<void*>(&sContactInfo));
		}
		return 0;
	}
	return 1;
}

int cScr_ContactTrigger::FindContactObject(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	ContactTrigger_Struct* pContactInfo = reinterpret_cast<ContactTrigger_Struct*>(pData);
	cScr_ContactTrigger* pContactTrigger = static_cast<cScr_ContactTrigger*>(pScript);
	if (pContactInfo->bOn)
		pContactTrigger->DoTurnOn(0);
	else
		pContactTrigger->DoTurnOff(0);
	return 0;
}

int cScr_ContactTrigger::FindContactArchetype(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	ContactTrigger_Struct* pContactInfo = reinterpret_cast<ContactTrigger_Struct*>(pData);
	cScr_ContactTrigger* pContactTrigger = static_cast<cScr_ContactTrigger*>(pScript);
	SService<IObjectSrv> pOS(g_pScriptManager);
	sLink sl;
	pLQ->Link(&sl);
	true_bool child_p;
	if (*(pOS->InheritsFrom(child_p, pContactInfo->iObject, sl.dest)))
	{
		if (pContactInfo->bOn)
			pContactTrigger->DoTurnOn(0);
		else
			pContactTrigger->DoTurnOff(0);
		return 0;
	}
	return 1;
}


/*************
 * TrapText
 */
long cScr_TrapText::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (bTurnOn)
	{
		if (ShowBook(m_iObjId, false))
		{
			char* psz = GetObjectParams(m_iObjId);
			if (psz)
			{
				ShowString(psz, -1001, 0);
				g_pMalloc->Free(psz);
			}
		}
	}
	else
	{
		ShowString("", 1);
	}
	return 0;
}

long cScr_TrapText::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrMsg MyMsg(*pMsg);
	MyMsg.message = "TurnOn";
	return ReceiveMessage(&MyMsg, pReply, kNoAction);
}


/****************
 * FocusPlaque
 */
long __stdcall cScr_FocusPlaque::OnFocus(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	ShowBook(m_iObjId, false);
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long __stdcall cScr_FocusPlaque::OnDeFocus(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		if ((static_cast<int>(mpTrapFlags) & kTrapFlagNoOff) == 0)
		{
			ShowString("", 1);
			return 0;
		}
	}
	else
	{
			ShowString("", 1);
			return 0;
	}
	return 1;
}


/********************
 * ContainerPhysics
 */
long __stdcall cScr_ContainerPhysics::OnSlain(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SInterface<IContainSys> pContainSys(g_pScriptManager);
	sContainIter* pItem = pContainSys->IterStart(m_iObjId);
	if (pItem)
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		SService<IObjectSrv> pObjSrv(g_pScriptManager);
		SService<IPhysSrv> pPhysSrv(g_pScriptManager);
		SService<IDataSrv> pDataSrv(g_pScriptManager);
		cScrVec vDims;
		if (pPropSrv->Possessed(m_iObjId, "PhysType"))
		{
			cMultiParm mpShape;
			pPropSrv->Get(mpShape, m_iObjId, "PhysType", "Type");
			switch (static_cast<int>(mpShape))
			{
			case kPMT_Sphere:
			case kPMT_SphereHat:
				pPropSrv->Get(mpShape, m_iObjId, "PhysDims", "Radius 1");
				if (mpShape.type == kMT_Float)
				{
					vDims.pos.x = static_cast<float>(mpShape);
					vDims.pos.y = static_cast<float>(mpShape);
					vDims.pos.z = static_cast<float>(mpShape);
				}
				break;
			case kPMT_OBB:
				pPropSrv->Get(mpShape, m_iObjId, "PhysDims", "Size");
				if (mpShape.type == kMT_Vector)
				{
					vDims.pos.x = static_cast<mxs_vector*>(mpShape)->x / 2;
					vDims.pos.y = static_cast<mxs_vector*>(mpShape)->y / 2;
					vDims.pos.z = static_cast<mxs_vector*>(mpShape)->z / 2;
					g_pMalloc->Free(static_cast<mxs_vector*>(mpShape));
				}
				break;
			default:
				break;
			}
		}
		cScrVec vVelocity;
		pPhysSrv->GetVelocity(m_iObjId, vVelocity);
		float fAbsorption = (static_cast<sSlayMsg*>(pMsg)->kind == -900) ? -0.1 : 0.6;
		vVelocity.pos.x *= fAbsorption;
		vVelocity.pos.y *= fAbsorption;
		vVelocity.pos.z *= fAbsorption;
		cScrVec vOffset;
		cScrVec vZero;
		while (1)
		{
			vOffset.pos.x = vDims.pos.x * pDataSrv->RandFltNeg1to1();
			vOffset.pos.y = vDims.pos.y * pDataSrv->RandFltNeg1to1();
			vOffset.pos.z = vDims.pos.z * pDataSrv->RandFltNeg1to1();
			pObjSrv->Teleport(pItem->Object, vOffset, vZero, m_iObjId);
			pContainSys->Remove(pItem->Object, 0);
			pPropSrv->Remove(pItem->Object, "PhysType");
			pPropSrv->Add(pItem->Object, "PhysType");
			pPhysSrv->SetVelocity(pItem->Object, vVelocity);
			if (!pContainSys->IterNext(pItem))
				break;
		}
		pContainSys->IterEnd(pItem);
	}
	return 0;
}


